/*
 * Copyright (C) 2013 Castrense Savojardo
 * 	Bologna Biocomputing Group
 * 	University of Bologna, Italy
 *  savojard@biocomp.unibo.it
 *
 * IO.c - This file is part of BCov
 *
 * BCov is a free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * BCov is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BCov; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 */

#include "IO.h"
#include "mem.h"

void copyright() {
	fprintf(stderr, "* BCov v1.0\n");
	fprintf(stderr, "* Copyright (C) 2013 Castrense Savojardo\n");
	fprintf(stderr, "*  Bologna Biocomputing Group\n");
	fprintf(stderr, "*  University of Bologna, Italy\n");
	fprintf(stderr, "*  savojard@biocomp.unibo.it\n\n");
}



void print_matrix(entry **M, unsigned int x, unsigned int y) {
	unsigned int i, j;
	for (i=0; i<x; i++) {
		for (j=0; j<y; j++) {
			printf("%1.10f ", M[i][j].score);
		}
		printf("\n");
	}
}

void fail(char *errstr)
{
    fprintf(stderr, "\n*** %s\n\n", errstr);
    exit(-1);
}

void print_verbose(char *msgstr, int verbose, int required) {
	if (verbose >= required) {
		printf("%s", msgstr);
	}
}

void printe(char *errstr)
{
    fprintf(stderr, "\n*** %s\n\n", errstr);

}

void usage() {
	fprintf(stderr, "Usage: bcov [options]\n\n");
	fprintf(stderr, "Options:\n\n");
	fprintf(stderr, "-a file:\tInput MSA file in the PSICOV format. A contact score matrix \n");
	fprintf(stderr, "        \tis internally computed using the PSICOV method (Jones et al., 2012).\n");
	fprintf(stderr, "        \tThis option conflicts with the -m option. REQUIRED if -m is not set\n\n");

	fprintf(stderr, "-m file:\tRead a pre-computed residue contact score matrix from file. This\n");
	fprintf(stderr, "        \toption conflicts with the -c option. REQUIRED if -a is not set\n\n");

	fprintf(stderr, "-s file:\tRead beta strand boundaries from file. REQUIRED.\n\n");
	fprintf(stderr, "-c file:\tWrite predicted beta contact map to file. REQUIRED.\n\n");
	fprintf(stderr, "-o file:\tWrite predicted beta strand pairing to file. REQUIRED.\n\n");

	fprintf(stderr, "-n nnn:\t\tMinimum sequence separation for parallel strand pairing.\n");
	fprintf(stderr, "       \t\t0 = no threshold. OPTIONAL, default: 6\n\n");
	fprintf(stderr, "-v nnn:\t\tVerbose level, 0/1 (default: 0).\n\n");
	exit(-1);
}

double ** readbinarymat(char * filename, unsigned int * l) {
	unsigned int i, j;
	double ** mat;
	FILE * fp;
	if((fp=fopen(filename,"r"))==NULL) return NULL;
	if(fscanf(fp,"%d\n",l)==EOF) return NULL;
	mat = allocmat(*l, *l, sizeof(double));
	while(fscanf(fp,"%d %d\n",&i,&j)!=EOF) {
		if (i < *l && j < *l) {
			mat[i][j] = mat[j][i] = 1;
		} else return NULL;
	}
	fclose(fp);
	return mat;
}

double ** readmat(char * filename, unsigned int * l) {
	unsigned int i, j;
	double ** mat;
	float v;
	FILE * fp;
	if((fp=fopen(filename,"r"))==NULL) return NULL;
	if(fscanf(fp,"%d\n",l)==EOF) return NULL;
	mat = allocmat(*l, *l, sizeof(double));
	while(fscanf(fp,"%d %d %f\n",&i,&j,&v)!=EOF) {
		if (i < *l && j < *l) {
			mat[i][j] = mat[j][i] = v;
		} else return NULL;
	}
	fclose(fp);
	return mat;
}

pair * readss(char * filename, unsigned int * ns, unsigned int * maxlen) {
	unsigned int i, j, k;
	pair * p;
	FILE * fp;
	if((fp=fopen(filename,"r"))==NULL) return NULL;
	if(fscanf(fp,"%d\n",ns)==EOF) return NULL;
	p = allocvec(*ns, sizeof(pair));
	*maxlen = 0;
	while(fscanf(fp,"%d %d %d\n",&k,&i,&j)!=EOF) {
		p[k].i = i;
		p[k].j = j;
		if (j - i + 1 > *maxlen) {
			*maxlen = j - i + 1;
		}

	}
	fclose(fp);
	return p;
}

void write_problem(FILE * ofp, double ** H, unsigned int n) {
	unsigned int i, j;

	fprintf(ofp, "data;\n");
	fprintf(ofp, "param N := %d;\n", n);
	fprintf(ofp, "param S :");
	for (i=0; i<n; i++) {
		fprintf(ofp, " %d", i+1);
	}
	fprintf(ofp, ":=\n");
	for (i=0; i<n; i++) {
		fprintf(ofp, "%d", i+1);
		for (j=0; j<n; j++) {
			fprintf(ofp, " %f", H[i][j]);
		}
		if (i==n-1) {
			fprintf(ofp, ";\n");
		}else { fprintf(ofp, "\n"); }
	}
}

void print_acc(double ** R, double ** P, unsigned int l, unsigned int tnbase) {
	int i, j, tp, fp, tn, fn;
	tp = fp = fn = 0;
	for (i=0; i<l-1; i++) {
		for (j=i+1; j<l; j++) {
			if (R[i][j] == P[i][j]) {
				tp += R[i][j];
			} else {
				fp += P[i][j];
				fn += (1 - P[i][j]);
			}
		}
	}
	tn = tnbase - fp - fn - tp;
	printf("TP %d\tTN %d\tFP %d\tFN %d\n", tp, tn, fp, fn);
}




