/*
 * Copyright (C) 2013 Castrense Savojardo
 * 	Bologna Biocomputing Group
 * 	University of Bologna, Italy
 *  savojard@biocomp.unibo.it
 *
 * main.c - This file is part of BCov
 *
 * BCov is a free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * BCov is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BCov; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include "mem.h"
#include "align.h"
#include "lp.h"
#include "IO.h"
#include "psicov.h"

#define MINSEP 6

void align_parallel(alignment * res,
					double ** S,
					pair ssi,
					pair ssj,
					unsigned int maxgaps,
					unsigned int type) {
	int i, j;
	double ** sub;
	unsigned int is = ssi.i, ie = ssi.j, js = ssj.i, je = ssj.j;

	sub = allocmat(ie - is + 1, je - js + 1, sizeof(double));
	for (i=is; i<= ie; i++) {
		for (j=js; j<=je; j++) {
			sub[i - is][j - js] = S[i][j];
		}
	}
	align_strands(res, sub, ie - is + 1, je - js + 1, maxgaps, type);
	freemat(sub);
}

void align_antiparallel(alignment * res,
						double ** S,
						pair ssi,
						pair ssj,
						unsigned int maxgaps,
						unsigned int type) {
	int i, j;
	double ** sub;
	unsigned int is = ssi.i, ie = ssi.j, js = ssj.i, je = ssj.j;

	sub = allocmat(ie - is + 1, je - js + 1, sizeof(double));
	for (i=is; i<= ie; i++) {
		for (j=je; j>=js; j--) {
			sub[i - is][je - j] = S[i][j];
		}
	}
	align_strands(res, sub, ie - is + 1, je - js + 1, maxgaps, type);
	freemat(sub);
}

unsigned int best_alignment(alignment * res,
							double ** S,
							pair ssi,
							pair ssj,
							unsigned int maxgaps,
							unsigned int type,
							unsigned int minsep) {
	unsigned int is = ssi.i, ie = ssi.j, js = ssj.i, je = ssj.j, d;
	alignment * parallel, * antiparallel;

	parallel = (alignment *) malloc(sizeof(alignment));
	align_parallel(parallel, S, ssi, ssj, maxgaps, type);
	d = PARALLEL;
	if (ie - is + 1 > 1 && je - js + 1 > 1) {
		antiparallel = (alignment *) malloc(sizeof(alignment));
		align_antiparallel(antiparallel, S, ssi, ssj, maxgaps, type);
		if (antiparallel->score >= parallel->score || js - ie - 1 < minsep ) {
			d = ANTIPARALLEL;
			copyaln(res, antiparallel);
		} else {
			copyaln(res, parallel);
		}
		freealn(antiparallel);
	} else {
		copyaln(res, parallel);
		d = BRIDGE;
	}

	freealn(parallel);
	return d;
}

int main(int argc, char ** argv) {
	int opt, ssflag, alnflag, matflag, oflag, omatflag, bridgeflag;
	unsigned int i, j, k, l, a, b, ns, nrs, maxlen, maxgaps, pairing_algo, alignment_algo, d, minsep, verbose = 0;
	int ** C;
	char * ofile, * alnfile, * mfile, * ssfile, * omapfile, * rmapfile, * rpairfile;
	double ** S, avesc, stdsc;
	double ** H, ** P, ** M;
	FILE * ofilep, * omapfilep;
	pair * ss;
	alignment * best;

	ssflag = alnflag = matflag = oflag = omatflag = 0;
	bridgeflag = BB;
	maxgaps = MAXBULGES;
	pairing_algo = MAXOV;
	alignment_algo = MAXOV;
	minsep = MINSEP;
	while ((opt = getopt(argc, argv, "c:s:m:o:a:n:v:")) >= 0) {
		switch (opt) {
			case 's':
				ssflag = 1;
				ssfile = strdup(optarg);
				break;
			case 'm':
				matflag = 1;
				mfile = strdup(optarg);
				break;
			case 'o':
				oflag = 1;
				ofile = strdup(optarg);
				break;
			case 'c':
				omatflag = 1;
				omapfile = strdup(optarg);
				break;
			case 'n':
				minsep = atoi(optarg);
				break;
			case 'a':
				alnflag = 1;
				alnfile = strdup(optarg);
				break;
			case 'v':
				verbose = atoi(optarg);
				break;
		}
	}

	copyright();

	if (matflag && alnflag) {
		printe("Options -m and -a are incompatible.");
		usage();
	}

	if ((! alnflag && ! matflag) || ! ssflag || ! omatflag || ! oflag) {
		printe("Some required arguments are not specified.");
		usage();
	}

	if((omapfilep=fopen(omapfile,"w"))==NULL) {
		fail("Error opening output contact map file.\n");
	}

	if((ofilep=fopen(ofile,"w"))==NULL) {
		fail("Error opening output pairing file.\n");
	}

	if (matflag) {
		print_verbose("** Reading residue contact score matrix from file...", verbose, 1);
		S = readmat(mfile, &l);
		print_verbose("done.\n", verbose, 1);
	}
	else {
		print_verbose("** Computing contact scores using PSICOV...", verbose, 0);
		fflush(stdout);
		psicov_args pargs;
		psicov_init(&pargs);
		S = psicov_sce(alnfile, &pargs, NULL, &l);
		print_verbose("done.\n", verbose, 0);
	}

	ss = readss(ssfile, &ns, &maxlen);
	H = allocmat(ns, ns, sizeof(double));

	print_verbose("** Computing strand pairing scores...", verbose, 0);
	avesc = 0.0;
	stdsc = 0.0;

	k = 0;
	for (i=0; i<ns; i++) {
		k += (ss[i].j - ss[i].i + 1 == 1) ? 1 : 0;
		for (j=i+1; j<ns; j++) {
			best = (alignment *) malloc(sizeof(alignment));
			best_alignment(best, S, ss[i], ss[j], maxgaps, pairing_algo, 0);
			H[i][j] = H[j][i] = best->score;
			avesc += H[i][j];
			stdsc += (H[i][j] * H[i][j]);
			free(best->pairs);
			free(best);
		}
	}

	C = allocmat(ns, ns, sizeof(int));

	stdsc = (1.0/(0.5*ns*(ns-1.0))) * sqrt((0.5*ns*(ns-1.0)) * stdsc - avesc * avesc);
	avesc = avesc / (0.5*ns*(ns-1.0));
	for (i=0; i<ns-1; i++) {
		for (j=i+1; j<ns; j++) {
			C[i][j] = C[j][i] = 1;
			if (stdsc != 0) {
				H[i][j] = (H[i][j] - avesc) / stdsc;
				H[j][i] = H[i][j];
			}
			if (ss[i].j == ss[i].i) {
				if (ss[j].j != ss[j].i && bridgeflag == BB && k > 1) {
					C[i][j] = C[j][i] = 0;
				}
			}
		}
	}
	print_verbose("done.\n", verbose, 0);
	print_verbose("** Solving integer program...", verbose, 0);

	P = allocmat(ns, ns, sizeof(double));
	solve_problem(H, P, C, ns);
	print_verbose("done.\n", verbose, 0);

	print_verbose("** Writing output files...", verbose, 1);
	fprintf(omapfilep, "%d\n", l);
	fprintf(ofilep, "%d\n", ns);

	for (i=0; i<ns; i++) {
		for (j=i+1; j<ns; j++) {
			if (P[i][j] > 0) {
				best = (alignment *) malloc(sizeof(alignment));
				d = best_alignment(best, S, ss[i], ss[j], maxgaps, alignment_algo, minsep);
				fprintf(ofilep, "%d %d %d\n", i, j, d);
				for (k=0; k<best->size; k++) {
					if (best->pairs[k].i != GAP && best->pairs[k].j != GAP) {
						a = best->pairs[k].i + ss[i].i;
						if (d == PARALLEL || d == BRIDGE) {
							b = best->pairs[k].j + ss[j].i;
						} else {
							b = ss[j].j - best->pairs[k].j;
						}
						fprintf(omapfilep, "%d %d\n", a, b);
					}
				}
				freealn(best);
			}
		}
	}
	print_verbose("done.\n", verbose, 1);

	freemat(H);
	freemat(S);
	freemat(P);
	freemat(C);
	free(ss);

	fclose(omapfilep);
	fclose(ofilep);

	return 0;
}


