/*
 * Copyright (C) 2013 Castrense Savojardo
 * 	Bologna Biocomputing Group
 * 	University of Bologna, Italy
 *  savojard@biocomp.unibo.it
 *
 * psicov.c - This file is part of BCov
 *
 * BCov is a free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * BCov is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BCov; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <unistd.h>

#include "mem.h"
#include "IO.h"
#include "psicov.h"

/* Convert AA letter to numeric code (0-21) */
int
aanum(int ch)
{
	const static int aacvs[] =
	{
			999, 0, 3, 4, 3, 6, 13, 7, 8, 9, 21, 11, 10, 12, 2,
			21, 14, 5, 1, 15, 16, 21, 19, 17, 21, 18, 6
	};

	return (isalpha(ch) ? aacvs[ch & 31] : 20);
}

/* Perform Cholesky decomposition on matrix */
int test_cholesky(double *a, const int n)
{
	int i, j, k;
	double sum;
	static double *diag;

	if (diag == NULL)
		diag = allocvec(n, sizeof(double));

	for (i=0; i<n; i++)
	{
		for (j=i; j<n; j++)
		{
			sum = a[i*n+j];

			for (k=i-1; k >= 0; k--)
				sum -= a[i*n+k]*a[j*n+k];

			if (i == j)
			{
				if (sum <= 0.0)
					return TRUE;

				diag[i] = sqrt(sum);
			}
			else
				a[j*n+i] = sum / diag[i];
		}
	}

	return FALSE;
}

struct sc_entry
{
	double sc;
	int i, j;
} *sclist;

/* Sort descending */
int cmpfn(const void *a, const void *b)
{
	if (((struct sc_entry *)a)->sc == ((struct sc_entry *)b)->sc)
		return 0;

	if (((struct sc_entry *)a)->sc < ((struct sc_entry *)b)->sc)
		return 1;

	return -1;
}

extern glasso_(int *, double *, double *, int *, int *, int *, int *, double *, int *, double *, double *, int *, double *, int *);


void psicov_init(psicov_args * pargs) {
	pargs->apcflg     = APCFLG;
	pargs->approxflg  = APPROXFLG;
	pargs->filtflg    = FILTFLG;
	pargs->idthresh   = IDTHRESH;
	pargs->maxgapf    = MAXGAPF;
	pargs->minseqsep  = MINSEQSEP;
	pargs->pseudoc    = PSEUDOC;
	pargs->rawscflg   = RAWSCFLG;
	pargs->rhodefault = RHO;
	pargs->shrinkflg  = SHRINKFLG;
	pargs->targfnzero = TARGFNZERO;
	pargs->thresh     = THRESH;
}

double ** psicov_sce(char * filename, psicov_args * pargs, char * blockfn, unsigned int * l)
{
	int a, b, i, j, k, seqlen, nids, s, nseqs, ncon;
	int opt, ndim, initflg=0, debugflg=0, diagpenflg=1;
	int maxit=10000, npair, nnzero, niter, jerr;
	unsigned int *wtcount, ccount[MAXSEQLEN];
	double del, **pcmat, *pcsum, pcmean, pc, trialrho;
	double sum, score, (**pab)[21][21], **pa, wtsum, lambda, smean, fnzero, lastfnzero, rfact, r2, scsum, scsumsq, mean, sd, zscore, ppv;
	double *weight, ** C;
	char buf[4096], seq[MAXSEQLEN], **aln;
	FILE *ifp;


	ifp = fopen(filename, "r");
	if (!ifp)
		fail("Unable to open alignment file!");

	for (nseqs=0;; nseqs++)
		if (!fgets(seq, MAXSEQLEN, ifp))
			break;

	aln = allocvec(nseqs, sizeof(char *));

	weight = allocvec(nseqs, sizeof(double));

	wtcount = allocvec(nseqs, sizeof(unsigned int));

	rewind(ifp);

	if (!fgets(seq, MAXSEQLEN, ifp))
		fail("Bad alignment file!");

	seqlen = strlen(seq)-1;

	if (!(aln[0] = malloc(seqlen)))
		fail("Out of memory!");

	for (j=0; j<seqlen; j++)
		aln[0][j] = aanum(seq[j]);

	for (i=1; i<nseqs; i++)
	{
		if (!fgets(seq, MAXSEQLEN, ifp))
			break;

		if (seqlen != strlen(seq)-1)
			fail("Length mismatch in alignment file!");

		if (!(aln[i] = malloc(seqlen)))
			fail("Out of memory!");

		for (j=0; j<seqlen; j++)
			aln[i][j] = aanum(seq[j]);
	}


	/* Calculate sequence weights */

	if (pargs->idthresh < 0.0)
	{
		double meanfracid = 0.0;

		for (i=0; i<nseqs; i++)
			for (j=i+1; j<nseqs; j++)
			{
				int nids;
				double fracid;

				for (nids=k=0; k<seqlen; k++)
					if (aln[i][k] == aln[j][k])
						nids++;

				fracid = (double)nids / seqlen;

				meanfracid += fracid;
			}

		meanfracid /= 0.5 * nseqs * (nseqs - 1.0);

		pargs->idthresh = 0.38 * 0.32 / meanfracid;
	}

	for (i=0; i<nseqs; i++)
		for (j=i+1; j<nseqs; j++)
		{
			int nthresh = seqlen * pargs->idthresh;

			for (k=0; nthresh > 0 && k<seqlen; k++)
				if (aln[i][k] != aln[j][k])
					nthresh--;

			if (nthresh > 0)
			{
				wtcount[i]++;
				wtcount[j]++;
			}
		}


	for (wtsum=i=0; i<nseqs; i++)
		wtsum += (weight[i] = 1.0 / (1 + wtcount[i]));

	/*
	if (wtsum < MINEFSEQS)
		fail("Sorry - not enough sequences or sequence diversity to proceed!\nNeff (%f) < MINEFSEQS (%d)\nIf you want to force a calculation, try reducing MINEFSEQS at your own risk.\n", wtsum, MINEFSEQS);
	*/
	pa = allocmat(seqlen, 21, sizeof(double));
	pab = allocmat(seqlen, seqlen, 21*21*sizeof(double));

	/* Calculate singlet frequencies with pseudocount */
	for (i=0; i<seqlen; i++)
	{
		for (a=0; a<21; a++)
			pa[i][a] = pargs->pseudoc;

		for (k=0; k<nseqs; k++)
		{
			a = aln[k][i];
			if (a < 21)
				pa[i][a] += weight[k];
		}

		for (a=0; a<21; a++)
			pa[i][a] /= pargs->pseudoc * 21.0 + wtsum;
	}

	/* Calculate pair frequencies with pseudocount */
	for (i=0; i<seqlen; i++)
	{
		for (j=i+1; j<seqlen; j++)
		{
			for (a=0; a<21; a++)
				for (b=0; b<21; b++)
					pab[i][j][a][b] = pargs->pseudoc / 21.0;

			for (k=0; k<nseqs; k++)
			{
				a = aln[k][i];
				b = aln[k][j];
				if (a < 21 && b < 21)
					pab[i][j][a][b] += weight[k];
			}

			for (a=0; a<21; a++)
				for (b=0; b<21; b++)
				{
					pab[i][j][a][b] /= pargs->pseudoc * 21.0 + wtsum;
					pab[j][i][b][a] = pab[i][j][a][b];

					//		    printf("%d/%d %d/%d %f %f %f %f\n", i+1, a, j+1, b, pab[i][j][a][b], pa[i][a] , pa[j][b], pab[i][j][a][b] - pa[i][a] * pa[j][b]);
				}
		}
	}

	for (i=0; i<seqlen; i++)
		for (a=0; a<21; a++)
			for (b=0; b<21; b++)
				pab[i][i][a][b] = (a == b) ? pa[i][a] : 0.0;

	double *cmat, *rho, *ww, *wwi, *tempmat;

	ndim = seqlen * 21;

	cmat = f_matrix_calloc(ndim, sizeof(double));
	tempmat = f_matrix_calloc(ndim, sizeof(double));

	/* Form the covariance matrix */
	for (i=0; i<seqlen; i++)
		for (j=0; j<seqlen; j++)
			for (a=0; a<21; a++)
				for (b=0; b<21; b++)
					if (i != j || a == b)
						cmat[(i*21+a) * ndim + j*21+b] = pab[i][j][a][b] - pa[i][a] * pa[j][b];

	freemat(pab);

	/* Shrink sample covariance matrix towards shrinkage target F = Diag(1,1,1,...,1) * smean */

	if (pargs->shrinkflg)
	{
		for (smean=i=0; i<ndim; i++)
			smean += cmat[i*ndim+i];

		smean /= (double)ndim;
		lambda = 0.1;

		for (;;)
		{
			memcpy(tempmat, cmat, ndim*ndim*sizeof(double));

			/* Test if positive definite using Cholesky decomposition */
			if (!test_cholesky(tempmat, ndim))
				break;

			for (i=0; i<seqlen; i++)
				for (j=0; j<seqlen; j++)
					for (a=0; a<21; a++)
						for (b=0; b<21; b++)
							if (i != j)
								cmat[(i*21+a)*ndim + j*21+b] *= 1.0 - lambda;
							else if (a == b)
								cmat[(i*21+a)*ndim + j*21+b] = smean * lambda + (1.0 - lambda) * cmat[(i*21+a)*ndim + j*21+b];
		}
	}

	rho = f_matrix_calloc(ndim, sizeof(double));
	ww = f_matrix_calloc(ndim, sizeof(double));
	wwi = f_matrix_calloc(ndim, sizeof(double));

	lastfnzero=0.0;

	/* Guess at a reasonable starting rho value if undefined */
	if (pargs->rhodefault < 0.0)
		trialrho = MAX(0.001, 1.0 / wtsum);
	else
		trialrho = pargs->rhodefault;

	rfact = 0.0;

	for (;;)
	{
		if (trialrho <= 0.0 || trialrho >= 1.0)
			fail("Sorry - failed to find suitable value for rho (0 < rho < 1)!");

		for (i=0; i<ndim; i++)
			for (j=0; j<ndim; j++)
				rho[i*ndim + j] = trialrho;

		for (i=0; i<seqlen; i++)
			for (j=0; j<seqlen; j++)
				for (a=0; a<21; a++)
					for (b=0; b<21; b++)
						if ((a != b && i == j) || pa[i][20] > pargs->maxgapf || pa[j][20] > pargs->maxgapf)
							rho[(i*21+a)*ndim + j*21+b] = 1e9;

		/* Mask out regions if block-out list provided */
		if (blockfn != NULL)
		{
			ifp = fopen(blockfn, "r");

			for (;;)
			{
				if (fscanf(ifp, "%d %d %lf", &i, &j, &score) != 3)
					break;

				for (a=0; a<21; a++)
					for (b=0; b<21; b++)
					{
						rho[((i-1)*21+a)*ndim + (j-1)*21+b] = score;
						rho[((j-1)*21+b)*ndim + (i-1)*21+a] = score;
					}
			}

			fclose(ifp);
		}

		/* All matrices are symmetric so no need to transpose before/after calling Fortran code */

		glasso_(&ndim, cmat, rho, & (pargs->approxflg), &initflg, &debugflg, &diagpenflg, &(pargs->thresh), &maxit, ww, wwi, &niter, &del, &jerr);

		if (pargs->targfnzero <= 0.0)
			break;

		for (npair=nnzero=i=0; i<ndim; i++)
			for (j=i+1; j<ndim; j++,npair++)
				if (wwi[i*ndim+j] != 0.0)
					nnzero++;

		fnzero = (double) nnzero / npair;

		//      printf("rho=%f fnzero = %f\n", trialrho, fnzero);

		/* Stop iterating if we have achieved the target sparsity level */
		if (fabs(fnzero - pargs->targfnzero)/pargs->targfnzero < 0.01)
			break;

		if (fnzero == 0.0)
		{
			/* As we have guessed far too high, halve rho and try again */
			trialrho *= 0.5;
			continue;
		}

		if (lastfnzero > 0.0 && fnzero != lastfnzero)
		{
			//	    printf("fnzero=%f lastfnzero=%f trialrho=%f oldtrialrho=%f\n", fnzero, lastfnzero, trialrho, trialrho/rfact);

			rfact = pow(rfact, log(pargs->targfnzero / fnzero) / log(fnzero / lastfnzero));

			//	    printf("New rfact = %f\n", rfact);
		}

		lastfnzero = fnzero;

		/* Make a small trial step in the appropriate direction */

		if (rfact == 0.0)
			rfact = (fnzero < pargs->targfnzero) ? 0.9 : 1.1;

		trialrho *= rfact;
	}

	free(rho);
	free(ww);

	/* Calculate background corrected scores using average product correction */

	pcmat = allocmat(seqlen, seqlen, sizeof(double));
	pcsum = allocvec(seqlen, sizeof(double));

	pcmean = 0.0;

	for (i=0; i<seqlen; i++)
		for (j=i+1; j<seqlen; j++)
		{
			for (pc=a=0; a<20; a++)
				for (b=0; b<20; b++)
					pc += fabs(wwi[(i*21+a)*ndim + j*21+b]);

			pcmat[i][j] = pcmat[j][i] = pc;
			pcsum[i] += pc;
			pcsum[j] += pc;

			pcmean += pc;
		}

	pcmean /= seqlen * (seqlen - 1) * 0.5;

	/* Build final list of predicted contacts */

	sclist = allocvec(seqlen * (seqlen - 1) / 2, sizeof(struct sc_entry));

	for (scsum=scsumsq=ncon=i=0; i<seqlen; i++)
		for (j=i+pargs->minseqsep; j<seqlen; j++)
			if (pcmat[i][j] > 0.0)
			{
				/* Calculate APC score */
				if (pargs->apcflg)
					sclist[ncon].sc = pcmat[i][j] - pcsum[i] * pcsum[j] / SQR(seqlen - 1.0) / pcmean;
				else
					sclist[ncon].sc = pcmat[i][j];
				scsum += sclist[ncon].sc;
				scsumsq += SQR(sclist[ncon].sc);
				sclist[ncon].i = i;
				sclist[ncon++].j = j;
			}

	qsort(sclist, ncon, sizeof(struct sc_entry), cmpfn);

	mean = scsum / ncon;
	sd = 1.25 * sqrt(scsumsq / ncon - SQR(mean)); /* Corrected for extreme-value bias */

	for (i=0; i<seqlen; i++)
		ccount[i] = 0;

	C = allocmat(seqlen, seqlen, sizeof(double));
	/* Print output in CASP RR format with optional PPV estimated from final Z-score */
	if (pargs->rawscflg)
		for (i=0; i<ncon; i++) {
			//printf("%d %d 0 8 %f\n", sclist[i].i+1, sclist[i].j+1, sclist[i].sc);
			C[sclist[i].i][sclist[i].j] = sclist[i].sc;
		} else {
			for (i=0; i<ncon; i++)
			{
				zscore = (sclist[i].sc - mean) / sd;
				ppv = 0.904 / (1.0 + 16.61 * exp(-0.8105 * zscore));
				if (ppv >= 0.5 || (!ccount[sclist[i].i] || !ccount[sclist[i].j]) || !pargs->filtflg)
				{
					//printf("%d %d 0 8 %f\n", sclist[i].i+1, sclist[i].j+1, ppv);
					C[sclist[i].i][sclist[i].j] = ppv;
					ccount[sclist[i].i]++;
					ccount[sclist[i].j]++;
				}
			}
		}
	*l = seqlen;
	return C;
}

double ** psicov_compute_profile(char * filename, double idthresh, int pseudoc) {

	int i, j, k, a, seqlen, nseqs;
	unsigned int *wtcount;
	double *weight, wtsum, **pa;
	char **aln, seq[MAXSEQLEN];
	FILE *ifp;

	ifp = fopen(filename, "r");
	if (!ifp)
		fail("Unable to open alignment file!");

	for (nseqs=0;; nseqs++)
		if (!fgets(seq, MAXSEQLEN, ifp))
			break;

	aln = allocvec(nseqs, sizeof(char *));

	weight = allocvec(nseqs, sizeof(double));

	wtcount = allocvec(nseqs, sizeof(unsigned int));

	rewind(ifp);

	if (!fgets(seq, MAXSEQLEN, ifp))
		fail("Bad alignment file!");

	seqlen = strlen(seq)-1;

	if (!(aln[0] = malloc(seqlen)))
		fail("Out of memory!");

	for (j=0; j<seqlen; j++)
		aln[0][j] = aanum(seq[j]);

	for (i=1; i<nseqs; i++)
	{
		if (!fgets(seq, MAXSEQLEN, ifp))
			break;

		if (seqlen != strlen(seq)-1)
			fail("Length mismatch in alignment file!");

		if (!(aln[i] = malloc(seqlen)))
			fail("Out of memory!");

		for (j=0; j<seqlen; j++)
			aln[i][j] = aanum(seq[j]);
	}


	/* Calculate sequence weights */

	if (idthresh < 0.0)
	{
		double meanfracid = 0.0;

		for (i=0; i<nseqs; i++)
			for (j=i+1; j<nseqs; j++)
			{
				int nids;
				double fracid;

				for (nids=k=0; k<seqlen; k++)
					if (aln[i][k] == aln[j][k])
						nids++;

				fracid = (double)nids / seqlen;

				meanfracid += fracid;
			}

		meanfracid /= 0.5 * nseqs * (nseqs - 1.0);

		idthresh = 0.38 * 0.32 / meanfracid;
	}

	for (i=0; i<nseqs; i++)
		for (j=i+1; j<nseqs; j++)
		{
			int nthresh = seqlen * idthresh;

			for (k=0; nthresh > 0 && k<seqlen; k++)
				if (aln[i][k] != aln[j][k])
					nthresh--;

			if (nthresh > 0)
			{
				wtcount[i]++;
				wtcount[j]++;
			}
		}


	for (wtsum=i=0; i<nseqs; i++)
		wtsum += (weight[i] = 1.0 / (1 + wtcount[i]));

	/*
	if (wtsum < MINEFSEQS)
		fail("Sorry - not enough sequences or sequence diversity to proceed!\nNeff (%f) < MINEFSEQS (%d)\nIf you want to force a calculation, try reducing MINEFSEQS at your own risk.\n", wtsum, MINEFSEQS);
	*/
	pa = allocmat(seqlen, 20, sizeof(double));

	/* Calculate singlet frequencies with pseudocount */
	for (i=0; i<seqlen; i++)
	{
		for (a=0; a<20; a++)
			pa[i][a] = pseudoc;

		for (k=0; k<nseqs; k++)
		{
			a = aln[k][i];
			if (a < 20)
				pa[i][a] += weight[k];
		}

		for (a=0; a<20; a++)
			pa[i][a] /= pseudoc * 20.0 + wtsum;
	}

	return pa;
}

